local util = require("util")
local crash_site = require("crash-site")

local created_items = function()
  return
  {
    ["iron-plate"] = 8,
    ["wood"] = 1,
    ["pistol"] = 1,
    ["firearm-magazine"] = 10,
    ["burner-mining-drill"] = 1,
    ["stone-furnace"] = 1
  }
end

local respawn_items = function()
  return
  {
    ["pistol"] = 1,
    ["firearm-magazine"] = 10
  }
end

local ship_items = function()
  return
  {
    ["firearm-magazine"] = 8
  }
end

local debris_items = function()
  return
  {
    ["iron-plate"] = 8
  }
end

local ship_parts = function()
  return crash_site.default_ship_parts()
end

local chart_starting_area = function()
  local r = global.chart_distance or 200
  local force = game.forces.player
  local surface = game.surfaces[1]
  local origin = force.get_spawn_position(surface)
  force.chart(surface, {{origin.x - r, origin.y - r}, {origin.x + r, origin.y + r}})
end

local tweak_map_gen = function()
  local surface = game.surfaces[1]
  local iron = "iron-ore"
  local copper = "copper-ore"
  local coal = "coal"
  local stone = "stone"
  local oil = "crude-oil"
  local uranium = "uranium-ore"
  local enemy = "enemy-base"
  local mgs = surface.map_gen_settings
  mgs.autoplace_controls[iron].size = "3"
  mgs.autoplace_controls[iron].frequency = "0.1"
  mgs.autoplace_controls[iron].richness = "0.2"
  mgs.autoplace_controls[copper].size = "3"
  mgs.autoplace_controls[copper].frequency = "0.1"
  mgs.autoplace_controls[copper].richness = "0.2"
  mgs.autoplace_controls[coal].size = "3"
  mgs.autoplace_controls[coal].frequency = "0.1"
  mgs.autoplace_controls[coal].richness = "0.2"
  mgs.autoplace_controls[stone].size = "3"
  mgs.autoplace_controls[stone].frequency = "0.1"
  mgs.autoplace_controls[stone].richness = "0.2"
  mgs.autoplace_controls[oil].size = "10"
  mgs.autoplace_controls[oil].frequency = "0.1"
  mgs.autoplace_controls[oil].richness = "0.01"
  mgs.autoplace_controls[uranium].size = "3"
  mgs.autoplace_controls[uranium].frequency = "0.1"
  mgs.autoplace_controls[uranium].richness = "0.2"
  mgs.autoplace_controls[enemy].size = "10"
  mgs.autoplace_controls[enemy].frequency = "0.1"
  surface.map_gen_settings = mgs
end


local on_player_created = function(event)
  local player = game.get_player(event.player_index)
  util.insert_safe(player, global.created_items)

  if not global.init_ran then

    --This is so that other mods and scripts have a chance to do remote calls before we do things like charting the starting area, creating the crash site, etc.
	--Debug Settings (disable)
	--game.difficulty_settings.technology_difficulty = 1
	--game.difficulty_settings.technology_price_multiplier = 0.001
	--Unit Group Settings
    game.map_settings.unit_group.max_gathering_unit_groups = 20
	game.map_settings.unit_group.max_unit_group_size = 300
	--Enemy Expansion Settings
	game.map_settings.enemy_expansion.enabled = false
	game.map_settings.enemy_expansion.max_expansion_distance = 0
	--game.map_settings.enemy_expansion.settler_group_min_size = 30
	--game.map_settings.enemy_expansion.settler_group_max_size = 50
	--game.map_settings.enemy_expansion.min_expansion_cooldown = 11000
	--game.map_settings.enemy_expansion.max_expansion_cooldown = 22000
    game.map_settings.enemy_expansion.friendly_base_influence_radius = 0
	--game.map_settings.enemy_expansion.other_base_coefficient = 30
	--game.map_settings.enemy_expansion.max_colliding_tiles_coefficient = 0.1
	--Enemy Evolution Settings
	game.map_settings.enemy_evolution.time_factor = 0.00002
	game.map_settings.enemy_evolution.destroy_factor = 0.002
	game.map_settings.enemy_evolution.pollution_factor = 0.0000012
	--Pollution Settings
	game.map_settings.pollution.ageing = 1
	game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.5
	--Day Settings
    player.surface.brightness_visual_weights = { 1 / 0.85, 1 / 0.85, 1 / 0.85 }
	player.surface.ticks_per_day = 200000
    player.surface.dawn = 0.80
    player.surface.dusk = 0.20
    player.surface.evening = 0.40
    player.surface.morning = 0.60
    player.surface.daytime = 0.80
	player.surface.freeze_daytime = false
	game.forces["player"].max_successful_attempts_per_tick_per_construction_queue = 6
	game.forces["player"].max_failed_attempts_per_tick_per_construction_queue = 3
--	game.forces["enemy"].set_ammo_damage_modifier("melee", 1)
--	game.forces["enemy"].set_ammo_damage_modifier("biological", 1)
--	game.forces["player"].set_turret_attack_modifier("flamethrower-turret", -0.5)
--	game.forces["player"].set_turret_attack_modifier("gun-turret", -0.5)
--	game.forces["player"].set_turret_attack_modifier("laser-turret", -0.5)
	--End
    global.init_ran = true

    chart_starting_area()
	tweak_map_gen()

    if not global.disable_crashsite then
      local surface = player.surface
      crash_site.create_crash_site(surface, {-5,-6}, util.copy(global.crashed_ship_items), util.copy(global.crashed_debris_items), util.copy(global.crashed_ship_parts))
      util.remove_safe(player, global.crashed_ship_items)
      util.remove_safe(player, global.crashed_debris_items)
      player.get_main_inventory().sort_and_merge()
      if player.character then
        player.character.destructible = false
      end
      global.crash_site_cutscene_active = true
      crash_site.create_cutscene(player, {-5, -4})
      return
    end

  end

  if not global.skip_intro then
    if game.is_multiplayer() then
      player.print(global.custom_intro_message or {"msg-intro"})
    else
      game.show_message_dialog{text = global.custom_intro_message or {"msg-intro"}}
    end
  end

end

local on_player_respawned = function(event)
  local player = game.get_player(event.player_index)
  util.insert_safe(player, global.respawn_items)
end

local on_research_finished = function(event)
--  if (event.research.name == "automation-2") then
--  game.map_settings.pollution.ageing = 0.5
--  game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.8
--  end
--  if (event.research.name == "oil-processing") then
--  game.map_settings.pollution.ageing = 0.4
--  game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.5
--  end
  if (event.research.name == "chemical-science-pack") then
--  game.map_settings.pollution.ageing = 0.3
  game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.3
  game.map_settings.enemy_evolution.time_factor = 0.00003
  game.map_settings.enemy_evolution.destroy_factor = 0.0025
  game.map_settings.enemy_evolution.pollution_factor = 0.0000014
  end
  if (event.research.name == "construction-robotics") then
  game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.2
  game.map_settings.enemy_evolution.time_factor = 0.00004
  game.map_settings.enemy_evolution.destroy_factor = 0.003
  game.map_settings.enemy_evolution.pollution_factor = 0.0000016
--  game.forces["enemy"].set_ammo_damage_modifier("melee", 1)
  end
  if (event.research.name == "rocket-control-unit") then
  --game.forces["enemy"].evolution_factor = 1
  game.map_settings.pollution.ageing = 0.85
  game.map_settings.pollution.enemy_attack_pollution_consumption_modifier = 0.05
  game.map_settings.enemy_evolution.time_factor = 0.0009
  game.map_settings.enemy_evolution.destroy_factor = 0.01
  game.map_settings.enemy_evolution.pollution_factor = 0.0000026
  game.forces["enemy"].set_ammo_damage_modifier("melee", 1)
  game.forces["enemy"].set_ammo_damage_modifier("biological", 1)
  game.forces["enemy"].set_gun_speed_modifier("melee", 1)
  end
--  if (event.research.name == "rocket-silo") then
--  game.forces["enemy"].set_gun_speed_modifier("melee", 1)
--  end
--  if (event.research.name == "physical-projectile-damage-1") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
--  if (event.research.name == "physical-projectile-damage-2") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
--  if (event.research.name == "physical-projectile-damage-3") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
--  if (event.research.name == "physical-projectile-damage-4") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
--  if (event.research.name == "physical-projectile-damage-5") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
--  if (event.research.name == "physical-projectile-damage-6") then
--  game.forces["player"].set_turret_attack_modifier("gun-turret", 0)
--  end
  if (event.research.name == "refined-flammables-1") then
  game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
  if (event.research.name == "refined-flammables-2") then
  game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
  if (event.research.name == "refined-flammables-3") then
  game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
  if (event.research.name == "refined-flammables-4") then
  game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
  if (event.research.name == "refined-flammables-5") then
 game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
  if (event.research.name == "refined-flammables-6") then
  game.forces["player"].set_turret_attack_modifier("flamethrower-turret", 0)
  end
end

local on_cutscene_waypoint_reached = function(event)
  if not global.crash_site_cutscene_active then return end
  if not crash_site.is_crash_site_cutscene(event) then return end

  local player = game.get_player(event.player_index)

  player.exit_cutscene()

  if not global.skip_intro then
    if game.is_multiplayer() then
      player.print(global.custom_intro_message or {"msg-intro"})
    else
      game.show_message_dialog{text = global.custom_intro_message or {"msg-intro"}}
    end
  end
end

local skip_crash_site_cutscene = function(event)
  if not global.crash_site_cutscene_active then return end
  if event.player_index ~= 1 then return end
  local player = game.get_player(event.player_index)
  if player.controller_type == defines.controllers.cutscene then
    player.exit_cutscene()
  end
end

local on_cutscene_cancelled = function(event)
  if not global.crash_site_cutscene_active then return end
  if event.player_index ~= 1 then return end
  global.crash_site_cutscene_active = nil
  local player = game.get_player(event.player_index)
  if player.gui.screen.skip_cutscene_label then
    player.gui.screen.skip_cutscene_label.destroy()
  end
  if player.character then
    player.character.destructible = true
  end
  player.zoom = 1.5
end

local on_player_display_refresh = function(event)
  crash_site.on_player_display_refresh(event)
end

local freeplay_interface =
{
  get_created_items = function()
    return global.created_items
  end,
  set_created_items = function(map)
    global.created_items = map or error("Remote call parameter to freeplay set created items can't be nil.")
  end,
  get_respawn_items = function()
    return global.respawn_items
  end,
  set_respawn_items = function(map)
    global.respawn_items = map or error("Remote call parameter to freeplay set respawn items can't be nil.")
  end,
  set_skip_intro = function(bool)
    global.skip_intro = bool
  end,
  get_skip_intro = function()
    return global.skip_intro
  end,
  set_custom_intro_message = function(message)
    global.custom_intro_message = message
  end,
  get_custom_intro_message = function()
    return global.custom_intro_message
  end,
  set_chart_distance = function(value)
    global.chart_distance = tonumber(value) or error("Remote call parameter to freeplay set chart distance must be a number")
  end,
  get_disable_crashsite = function()
    return global.disable_crashsite
  end,
  set_disable_crashsite = function(bool)
    global.disable_crashsite = bool
  end,
  get_init_ran = function()
    return global.init_ran
  end,
  get_ship_items = function()
    return global.crashed_ship_items
  end,
  set_ship_items = function(map)
    global.crashed_ship_items = map or error("Remote call parameter to freeplay set created items can't be nil.")
  end,
  get_debris_items = function()
    return global.crashed_debris_items
  end,
  set_debris_items = function(map)
    global.crashed_debris_items = map or error("Remote call parameter to freeplay set respawn items can't be nil.")
  end,
  get_ship_parts = function()
    return global.crashed_ship_parts
  end,
  set_ship_parts = function(parts)
    global.crashed_ship_parts = parts or error("Remote call parameter to freeplay set ship parts can't be nil.")
  end
}

if not remote.interfaces["freeplay"] then
  remote.add_interface("freeplay", freeplay_interface)
end

local is_debug = function()
  local surface = game.surfaces.nauvis
  local map_gen_settings = surface.map_gen_settings
  return map_gen_settings.width == 50 and map_gen_settings.height == 50
end

local freeplay = {}

freeplay.events =
{
  [defines.events.on_player_created] = on_player_created,
  [defines.events.on_player_respawned] = on_player_respawned,
  [defines.events.on_cutscene_waypoint_reached] = on_cutscene_waypoint_reached,
  ["crash-site-skip-cutscene"] = skip_crash_site_cutscene,
  [defines.events.on_player_display_resolution_changed] = on_player_display_refresh,
  [defines.events.on_player_display_scale_changed] = on_player_display_refresh,
  [defines.events.on_cutscene_cancelled] = on_cutscene_cancelled,
  [defines.events.on_research_finished] = on_research_finished
}

freeplay.on_configuration_changed = function()
  global.created_items = global.created_items or created_items()
  global.respawn_items = global.respawn_items or respawn_items()
  global.crashed_ship_items = global.crashed_ship_items or ship_items()
  global.crashed_debris_items = global.crashed_debris_items or debris_items()
  global.crashed_ship_parts = global.crashed_ship_parts or ship_parts()

  if not global.init_ran then
    -- migrating old saves.
    global.init_ran = #game.players > 0
  end
end

freeplay.on_init = function()
  global.created_items = created_items()
  global.respawn_items = respawn_items()
  global.crashed_ship_items = ship_items()
  global.crashed_debris_items = debris_items()
  global.crashed_ship_parts = ship_parts()

  if is_debug() then
    global.skip_intro = true
    global.disable_crashsite = true
  end

end

return freeplay
